#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-26
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from .ln2 import LN2Handler
from .thermistor import ThermistorHandler
from .valve import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-26
# @Filename: valve.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from datetime import UTC, datetime

from typing import TYPE_CHECKING, Any, Optional

from rich.progress import TaskID

from lvmopstools.clu import CluClient

from lvmcryo.config import get_internal_config
from lvmcryo.handlers.thermistor import ThermistorHandler
from lvmcryo.tools import (
    cancel_task,
    get_fake_logger,
)


if TYPE_CHECKING:
    from sdsstools.configuration import Configuration


__all__ = [
    "ValveHandler",
    "valve_on_off",
    "cancel_nps_threads",
    "close_all_valves",
    "outlet_info",
]


async def outlet_info(actor: str, outlet: str) -> dict[str, Any]:
    """Retrieves outlet information from the NPS."""

    async with CluClient() as client:
        cmd = await client.send_command(actor, f"status {outlet}")
        if cmd.status.did_fail:
            raise RuntimeError(f"Command '{actor} status {outlet}' failed.")

    return cmd.replies.get("outlet_info")


async def valve_on_off(
    actor: str,
    outlet_name: str,
    on: bool,
    timeout: float | None = None,
    use_script: bool = True,
    dry_run: bool = False,
) -> int | None:
    """Turns a valve on/off.

    Parameters
    ----------
    actor
        The NPS actor that commands the valve.
    outlet_name
        The name of the outlet to which the valve is connected.
    on
        Whether to turn the valve on or off.
    timeout
        Time, in seconds, after which the valve will be turned off.
    use_script
        If ``timeout`` is defined, the user script ``cycle_with_timeout``
        (must be defined in the NPS) to set a timeout after which the valve
        will be turned off. With ``use_script=False``, a ``timeout`` will
        block until the timeout is reached.
    dry_run
        Does not send the command to the NPS.

    Returns
    -------
    thread_id
        If ``use_script=True``, the thread number of the user script that
        was started. Otherwise `None`.

    """

    is_script: bool = False

    if on is True and isinstance(timeout, (int, float)) and use_script is True:
        # First we need to get the outlet number.
        info = await outlet_info(actor, outlet_name)
        id_ = info["id"]

        command_string = f"scripts run cycle_with_timeout {id_} {timeout}"
        is_script = True

    else:
        if on is False or timeout is None:
            command_string = f"{'on' if on else 'off'} {outlet_name}"
        else:
            command_string = f"on --off-after {timeout} {outlet_name}"

    if dry_run:
        return 0 if is_script else None

    async with CluClient() as client:
        command = await client.send_command(actor, command_string)
        if command.status.did_fail:
            raise RuntimeError(f"Command '{actor} {command_string}' failed")

    if is_script:
        script_data = command.replies.get("script")
        return script_data["thread_id"]

    return


async def cancel_nps_threads(actor: str, thread_id: int | None = None):
    """Cancels a script thread in an NPS.

    Parameters
    ----------
    actor
        The name of the NPS actor to command.
    thread_id
        The thread ID to cancel. If `None`, all threads in the NPS will be cancelled.

    """

    command_string = f"scripts stop {thread_id if thread_id else ''}"

    async with CluClient() as client:
        await client.send_command(actor, command_string)


async def close_all_valves(config: Configuration | None = None, dry_run: bool = False):
    """Closes all the outlets."""

    config = config or get_internal_config()
    valve_info = config["valve_info"]

    await asyncio.gather(
        *[
            valve_on_off(
                valve_info[valve]["actor"],
                valve_info[valve]["outlet"],
                False,
                dry_run=dry_run,
            )
            for valve in valve_info
        ]
    )


if TYPE_CHECKING:
    from lvmcryo.tools import TimerProgressBar


@dataclass
class ValveHandler:
    """Handles a valve, including opening and closing, timeouts, and thermistors.

    Parameters
    ----------
    valve
        The name of the valve. Additional information such as thermistor channel,
        actor names, etc. are determined from the configuration file if not
        explicitely provided.
    actor
        The NPS actor that command the valve.
    outlet
        The outlet name of the valve in the actor.
    thermistor_info
        Configuration data for the thermistor.
    progress_bar
        Progress bar instance used to display progress.
    dry_run
        Does not actually operate the valves.

    """

    valve: str
    actor: str
    outlet: str
    thermistor_info: dict[str, Any] | None = None
    progress_bar: Optional[TimerProgressBar] = None
    log: logging.Logger = field(default_factory=get_fake_logger)
    dry_run: bool = False

    def __post_init__(self):
        self._thread_id: int | None = None
        self._progress_bar_id: TaskID | None = None

        self._monitor_task: asyncio.Task | None = None
        self._timeout_task: asyncio.Task | None = None

        self.open_time: datetime | None = None
        self.close_time: datetime | None = None

        self.event = asyncio.Event()
        self.active: bool = False
        self.timed_out: bool = False

        self.thermistor: ThermistorHandler | None = None

    async def check(self):
        """Check the connection to the NPS."""

        try:
            info = await outlet_info(self.actor, self.outlet)
            assert isinstance(info, dict), "Invalid outlet info."
            assert not info["state"], "Valve is already open."
        except Exception as err:
            self.log.error(f"Error connecting to NPS: {err}")
            return False

        return True

    async def start_fill(
        self,
        min_open_time: float = 0.0,
        max_open_time: float | None = None,
        use_thermistor: bool = True,
        close_on_active: bool = True,
    ):
        """Starts a fill.

        Parameters
        ----------
        min_open_time
            Minimum time to keep the valve open.
        max_open_time
            The maximum time to keep the valve open. If ``None``, a default
            value is used.
        use_thermistor
            Whether to use the thermistor to close the valve. If ``True`` and
            ``fill_time`` is not ``None``, ``fill_time`` become the maximum
            open time.
        close_on_active
            If ``use_thermistor=True``, closes the valve when the thermistor becomes
            active. Otherwise just blocks while the thermistor is inactive.

        """

        if max_open_time is None:
            # Some hardcoded hard limits.
            if self.valve == "purge":
                max_open_time = 2000
            else:
                max_open_time = 600

            self.log.warning(
                f"No max_open_time provided for valve {self.valve!r}. "
                f"Using default of {max_open_time} seconds."
            )

        await self._set_state(True, timeout=max_open_time, use_script=True)

        if use_thermistor and self.thermistor_info:
            thermistor_channel = self.thermistor_info.pop("channel", self.valve)
            self.thermistor_info["close_valve"] = close_on_active
            self.thermistor = ThermistorHandler(
                self,
                channel=thermistor_channel,
                **self.thermistor_info,
            )
            self.thermistor.min_open_time = min_open_time
            self._monitor_task = asyncio.create_task(self.thermistor.start_monitoring())

        if self.progress_bar:
            if self.valve.lower() == "purge":
                initial_description = "Purge in progress ..."
                complete_description = "Purge complete"
            else:
                initial_description = "Fill in progress ..."
                complete_description = "Fill complete"

            self._progress_bar_id = await self.progress_bar.add_timer(
                max_open_time,
                label=self.valve,
                initial_description=initial_description,
                complete_description=complete_description,
            )

        await asyncio.sleep(2)
        self._timeout_task = asyncio.create_task(self._schedule_timeout(max_open_time))

        self.active = True

        self.event.clear()
        await self.event.wait()

    async def _schedule_timeout(self, timeout: float):
        """Schedules a task to cancel the fill after a timeout."""

        await asyncio.sleep(timeout)
        await self.finish(did_timeout=True)

    async def finish(self, close_valve: bool = True, did_timeout: bool = False):
        """Finishes the fill, closing the valve."""

        if close_valve:
            await self._set_state(False, did_timeout=did_timeout)

        self._monitor_task = await cancel_task(self._monitor_task)
        self._timeout_task = await cancel_task(self._timeout_task)

        if self._progress_bar_id is not None and self.progress_bar is not None:
            await self.progress_bar.stop_timer(self._progress_bar_id)
            self._progress_bar_id = None

        self.active = False

        if not self.event.is_set():
            self.event.set()

    async def _set_state(
        self,
        on: bool,
        use_script: bool = True,
        timeout: float | None = None,
        did_timeout: bool = False,
    ):
        """Sets the state of the valve.

        Parameters
        ----------
        on
            Whether to open or close the valve.
        use_script
            Whether to use the ``cycle_with_timeout`` script to set a timeout
            after which the valve will be closed.
        timeout
            Maximum open valve time.
        did_timeout
            Whether the valve was closed because of a timeout. Only affects the
            log message.

        """

        assert self.actor is not None

        # If there's already a thread running for this valve, we cancel it.
        if self._thread_id is not None and not self.dry_run:
            await cancel_nps_threads(self.actor, self._thread_id)

        thread_id = await valve_on_off(
            self.actor,
            self.valve,
            on,
            timeout=timeout,
            use_script=use_script,
            dry_run=self.dry_run,
        )

        if thread_id is not None:
            self._thread_id = thread_id

        if on:
            if thread_id is not None:
                self.log.info(
                    f"Valve {self.valve!r} was open with timeout={timeout} "
                    f"(thread_id={thread_id})."
                )
            else:
                self.log.info(f"Valve {self.valve!r} was open.")

            self.open_time = datetime.now(UTC)

        else:
            if did_timeout:
                self.log.warning(f"Valve {self.valve!r} was closed due to timeout.")
                self.timed_out = True
            else:
                self.log.info(f"Valve {self.valve!r} was closed.")

            self.close_time = datetime.now(UTC)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-26
# @Filename: ln2.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import datetime
import logging
from dataclasses import dataclass, field

from typing import Any, Callable, Coroutine, Literal, NoReturn, overload

import sshkeyboard
from pydantic import BaseModel, field_serializer
from rich import box
from rich.align import Align
from rich.console import Console
from rich.markup import render
from rich.panel import Panel

from lvmopstools.devices.specs import spectrograph_pressures, spectrograph_temperatures
from lvmopstools.devices.thermistors import read_thermistors
from sdsstools.utils import GatheringTaskGroup

from lvmcryo.config import ValveConfig, get_internal_config
from lvmcryo.handlers.thermistor import ThermistorMonitor
from lvmcryo.handlers.valve import ValveHandler
from lvmcryo.tools import TimerProgressBar, cancel_task, get_fake_logger, o2_alert


def convert_datetime_to_iso_8601_with_z_suffix(dt: datetime.datetime) -> str:
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")


def get_now():
    """Returns a UTC datetime for now."""

    return datetime.datetime.now(datetime.UTC)


class EventDict(BaseModel):
    """Dictionary of events."""

    start_time: datetime.datetime = field(default_factory=get_now)
    end_time: datetime.datetime | None = None
    purge_start: datetime.datetime | None = None
    purge_complete: datetime.datetime | None = None
    fill_start: datetime.datetime | None = None
    fill_complete: datetime.datetime | None = None
    fail_time: datetime.datetime | None = None
    abort_time: datetime.datetime | None = None

    @field_serializer("*")
    def serialize_dates(self, value: datetime.datetime | None) -> str | None:
        if value is None:
            return None
        return convert_datetime_to_iso_8601_with_z_suffix(value)


def get_valve_info():
    """Returns the valve information from the configuration file."""

    internal_config = get_internal_config()

    return {
        valve: ValveConfig(**data)
        for valve, data in internal_config["valve_info"].items()
    }


@dataclass
class LN2Handler:
    """The main LN2 purge/fill handlerclass.

    Parameters
    ----------
    cameras
        List of cameras to fill. Defaults to all the cameras.
    purge_valve
        The name of the purge valve.
    interactive
        Whether to show interactive features.
    log
        The logger instance. A new logger is created if not provided.
    valve_info
        A dictionary of valve name to actor and outlet name. If not provided,
        the internal configuration will be used.
    dry_run
        Does not actually operate the valves.
    alerts_route
        The API route to query system alerts.

    """

    cameras: list[str]
    purge_valve: str = "purge"
    interactive: bool = False
    log: logging.Logger = field(default_factory=get_fake_logger)
    valve_info: dict[str, ValveConfig] = field(default_factory=get_valve_info)
    dry_run: bool = False
    alerts_route: str | None = "http://lvm-hub.lco.cl:8090/api/alerts"

    def __post_init__(self):
        if self.interactive:
            console = getattr(self.log, "rich_console", None)
            self._progress_bar = TimerProgressBar(console)
            self.console = self._progress_bar.console
        else:
            self._progress_bar = None
            self.console = Console()

        self.valve_handlers: dict[str, ValveHandler] = {}
        for camera in self.cameras + [self.purge_valve]:
            if camera not in self.valve_info:
                raise ValueError(f"Cannot find valve info for {camera!r}.")

            actor = self.valve_info[camera].actor
            outlet = self.valve_info[camera].outlet
            thermistor = self.valve_info[camera].thermistor

            self.valve_handlers[camera] = ValveHandler(
                camera,
                actor,
                outlet,
                thermistor_info=thermistor.model_dump() if thermistor else None,
                progress_bar=self._progress_bar,
                log=self.log,
                dry_run=self.dry_run,
            )

        self._alerts_monitor_task: asyncio.Task | None = asyncio.create_task(
            self.monitor_alerts()
        )

        self.event_times = EventDict()

        self.failed: bool = False
        self.aborted: bool = False
        self.error: str | None = None

    def get_specs(self):
        """Returns a list of spectrographs being handled."""

        assert self.cameras is not None

        specs: set[str] = set([])
        for camera in self.cameras:
            cam_id = camera[-1]
            specs.add(f"sp{cam_id}")

        return specs

    async def check(
        self,
        max_pressure: float | None = None,
        max_temperature: float | None = None,
        check_thermistors: bool = True,
    ):
        """Checks if the pressure and temperature are in the allowed range.

        If ``max_temperature`` or ``max_pressure`` are not provided, the values
        from the configuration file are used.

        Parameters
        ----------
        max_pressure
            The maximum pressure to allow.
        max_temperature
            The maximum temperature to allow.
        check_thermistors
            Checks that the thermistors are reading correctly.

        Raises
        ------
        RuntimeError
            If any of the checks fails.

        """

        # Check O2 alarms.
        if not self.alerts_route:
            self.log.warning("No alerts route provided. Not checking O2 alarms.")
        else:
            try:
                self.log.info("Checking for O2 alarms ...")
                if await o2_alert(self.alerts_route):
                    self.fail("O2 alarm detected.")
                else:
                    self.log.debug("No O2 alarms reported.")
            except Exception as err:
                self.fail(f"Error checking O2 alarms: {err}")

        # Check connection to the NPS outlets
        try:
            self.log.info("Checking connection to NPS outlets ...")
            for valve in self.valve_handlers:
                if not await self.valve_handlers[valve].check():
                    raise RuntimeError(f"valve {valve} failed or did not reply.")
        except Exception as err:
            self.fail(f"Failed checking connection to NPS outlets: {err}")

        if max_temperature is not None:
            self.log.info("Checking LN2 temperatures ...")
            try:
                spec_temperatures = await spectrograph_temperatures()
            except Exception as err:
                self.fail(f"Failed reading spectrograph temperatures: {err}")
            else:
                for camera in self.cameras:
                    ln2_temp = spec_temperatures[f"{camera}_ln2"]

                    if ln2_temp is None:
                        self.fail(f"Failed retrieving {camera!r} temperature.")

                    if ln2_temp > max_temperature:
                        self.fail(
                            f"LN2 temperature for camera {camera} is {ln2_temp:.1f} C "
                            f"which is above the maximum allowed temperature "
                            f"({max_temperature:.1f} C)."
                        )

        if max_pressure is not None:
            self.log.info("Checking pressures ...")
            try:
                spec_pressures = await spectrograph_pressures()
            except Exception as err:
                self.fail(f"Failed reading spectrograph pressures: {err}")
            else:
                for camera in self.cameras:
                    pressure = spec_pressures[camera]

                    if pressure is None:
                        self.fail(f"Failed retrieving {camera!r} pressure.")

                    if pressure > max_pressure:
                        self.fail(
                            f"Pressure for camera {camera} is {pressure} Torr "
                            f"which is above the maximum allowed pressure "
                            f"({max_pressure} Torr)."
                        )

        if check_thermistors:
            self.log.info("Checking thermistors ...")

            try:
                thermistors = await read_thermistors()
            except Exception as err:
                self.fail(f"Failed reading thermistors: {err}")

            for valve in self.valve_handlers:
                thermistor_info = self.valve_info[valve].thermistor

                if thermistor_info is None:
                    self.log.warning(f"Cannot check thermistor for {valve}.")
                    continue

                if thermistor_info.disabled:
                    self.log.warning(f"Thermistor for {valve} is disabled.")
                    continue

                if thermistor_info.channel is None:
                    self.fail(f"Thermistor channel for {valve} not defined.")

                thermistor_value = thermistors[thermistor_info.channel]
                if thermistor_value is True:
                    self.fail(f"Thermistor for valve {valve} is active.")

        self.log.info("All pre-fill checks passed.")

        return True

    async def purge(
        self,
        purge_valve: str | None = None,
        use_thermistor: bool = True,
        min_purge_time: float | None = None,
        max_purge_time: float | None = None,
        prompt: bool | None = None,
        preopen_cb: Callable[[], Coroutine | Any] | None = None,
    ):
        """Purges the system.

        Parameters
        ----------
        purge_valve
            The name of the purge valve. If ``None``, uses the instance default.
        use_thermistor
            Whether to use the thermistor to close the valve.
        min_purge_time
            The minimum time to keep the purge valve open. Only relevant if
            using the thermistor.
        max_purge_time
            The maximum time to keep the purge valve open. If
            ``use_thermistor=False`` this is effectively the purge time unless
            ``prompt=True`` and the purge is cancelled before reaching the
            timeout.
        prompt
            Whether to show a prompt to stop or cancel the purge. If ``None``,
            determined from the instance ``interactive`` attribute.
        preopen_cb
            A callback to run before opening the purge valve.

        """

        if purge_valve is None:
            purge_valve = self.purge_valve

        valve_handler = self.valve_handlers[purge_valve]

        self.event_times.purge_start = get_now()

        self.log.info(
            f"Beginning purge using valve {valve_handler.valve!r} with "
            f"use_thermistor={use_thermistor}, min_open_time={min_purge_time}, "
            f"max_purge_time={max_purge_time}."
        )

        prompt = prompt if prompt is not None else self.interactive
        if prompt:
            self._kb_monitor(action="purge")

        try:
            if preopen_cb:
                if asyncio.iscoroutinefunction(preopen_cb):
                    await preopen_cb()
                else:
                    preopen_cb()

            await valve_handler.start_fill(
                min_open_time=min_purge_time or 0.0,
                max_open_time=max_purge_time,
                use_thermistor=use_thermistor,
            )

            if not self.aborted:
                self.log.info("Purge complete.")

        except Exception:
            self.fail()
            raise

        finally:
            self.event_times.purge_complete = get_now()
            if prompt:
                sshkeyboard.stop_listening()
                await asyncio.sleep(1)

            thermistor_monitor = ThermistorMonitor()  # Singleton
            thermistor_monitor.stop()

    async def fill(
        self,
        cameras: list[str] | None = None,
        use_thermistors: bool = True,
        require_all_thermistors: bool = False,
        min_fill_time: float | None = None,
        max_fill_time: float | None = None,
        prompt: bool | None = None,
        preopen_cb: Callable[[], Coroutine | Any] | None = None,
    ):
        """Fills the selected cameras.

        Parameters
        ----------
        cameras
            The list of cameras to fill. If not provided, defaults to the values
            used to instantiate the `.LN2Handler` instance.
        use_thermistors
            Whether to use the thermistors to close the valves.
        require_all_thermistors
            Whether to require all thermistors to be active before closing the
            valves. If `False`, valves as closed as their thermistors become active.
        min_fill_time
            The minimum time to keep the fill valves open. Only relevant if
            using the thermistors.
        max_fill_time
            The maximum time to keep the fill valves open. If
            ``use_thermistor=None`` this is effectively the fill time unless
            ``prompt=True`` and the fills are cancelled before reaching the
            timeout.
        prompt
            Whether to show a prompt to stop or cancel the fill. If ``None``,
            determined from the instance ``interactive`` attribute.
        preopen_cb
            A callback to run before opening the fill valves.

        """

        cameras = cameras or self.cameras
        if cameras is None or len(cameras) == 0:
            self.fail("No cameras selected for filling.")

        fill_tasks: list[Coroutine] = []

        for camera in cameras:
            try:
                valve_handler = self.valve_handlers[camera]
            except KeyError:
                self.fail(f"Unable to find valve for camera {camera!r}.")

            fill_tasks.append(
                valve_handler.start_fill(
                    min_open_time=min_fill_time or 0.0,
                    max_open_time=max_fill_time,
                    use_thermistor=use_thermistors,
                    close_on_active=not require_all_thermistors,
                )
            )

        self.event_times.fill_start = get_now()

        self.log.info(
            f"Beginning fill on cameras {cameras!r} with "
            f"use_thermistors={use_thermistors}, min_open_time={min_fill_time}, "
            f"max_fill_time={max_fill_time}."
        )

        prompt = prompt if prompt is not None else self.interactive
        if prompt:
            self._kb_monitor(action="fill")

        try:
            if preopen_cb:
                if asyncio.iscoroutinefunction(preopen_cb):
                    await preopen_cb()
                else:
                    preopen_cb()

            await asyncio.gather(*fill_tasks)

            if not self.aborted:
                self.log.info("Fill complete.")

        except Exception:
            self.fail()
            raise

        finally:
            if use_thermistors and require_all_thermistors:
                # If we are waiting for all thermistors to be active,
                # we need to close all valves now.
                self.log.info("Closing all valves.")
                async with GatheringTaskGroup() as group:
                    for valve_handler in self.valve_handlers.values():
                        group.create_task(valve_handler._set_state(False))

            self.event_times.fill_complete = get_now()
            if prompt:
                sshkeyboard.stop_listening()
                await asyncio.sleep(1)

            thermistor_monitor = ThermistorMonitor()  # Singleton
            thermistor_monitor.stop()

    def get_valve_times(
        self,
        as_string: bool = False,
    ) -> dict[str, dict[str, str | datetime.datetime | bool | None]]:
        """Returns a dictionary of open/close times for the valves."""

        result: dict[str, dict[str, str | datetime.datetime | bool | None]] = {}

        for valve in self.valve_handlers:
            result[valve] = {
                "open_time": None,
                "close_time": None,
                "timed_out": False,
                "thermistor_first_active": None,
            }

            open_time = self.valve_handlers[valve].open_time
            close_time = self.valve_handlers[valve].close_time

            if open_time is not None:
                result[valve]["open_time"] = (
                    open_time.isoformat() if as_string else open_time
                )

            if close_time is not None:
                result[valve]["close_time"] = (
                    close_time.isoformat() if as_string else close_time
                )

            thermistor = self.valve_handlers[valve].thermistor
            if thermistor is not None and thermistor.first_active is not None:
                result[valve]["thermistor_first_active"] = (
                    thermistor.first_active.isoformat()
                    if as_string
                    else thermistor.first_active
                )

            result[valve]["timed_out"] = self.valve_handlers[valve].timed_out

        return result

    def _kb_monitor(self, action: str = "fill"):
        """Monitors the keyboard and cancels/aborts the fill.."""

        async def monitor_keys(key: str):
            """Parses a pressed key and cancels/aborts the fills."""

            if key not in ["x", "X", "enter"]:
                return

            if key == "x" or key == "X":
                self.log.warning("Aborting now.")
                # No not raise an alert here.
                await self.abort(
                    error="Aborted by user.",
                    close_valves=True,
                    raise_error=False,
                )

            elif key == "enter":
                self.log.warning("Finishing purge/fill.")
                await self.close_valves(only_active=True)

        self.console.print(
            Panel(
                Align(
                    render(
                        f'Press [green]"enter"[/] to finish the {action} '
                        'or [green]"x"[/] to abort.'
                    ),
                    "center",
                ),
                box=box.HEAVY,
                border_style="green",
            )
        )

        # No need to store this task. It will be automatically done when
        # sshkeyboard.stop_listening() is called.
        asyncio.create_task(
            sshkeyboard.listen_keyboard_manual(
                on_press=monitor_keys,
                sleep=0.1,
            )
        )

    async def monitor_alerts(self):
        """Monitors the system alerts and aborts the fill if necessary."""

        if not self.alerts_route:
            self.log.warning("No alerts route provided. Not monitoring alerts.")
            return

        n_failed: int = 0

        while True:
            try:
                if await o2_alert(self.alerts_route):
                    await self.abort(
                        error="O2 alarm detected: closing valves and aborting.",
                        close_valves=True,
                        raise_error=False,
                    )

                n_failed = 0

            except Exception as ee:
                self.log.warning(f"Error reading alerts: {ee}")
                n_failed += 1

            if n_failed >= 10:
                await self.abort(
                    error="Too many errors reading alerts. Aborting.",
                    close_valves=True,
                    raise_error=False,
                )

            await asyncio.sleep(3)

    async def close_valves(self, only_active: bool = True):
        """Cancels ongoing fills and closes the valves.

        If ``only_active=True`` only active valves will be closed. Otherwise
        closes all valves.

        """

        tasks: list[Coroutine] = []

        for valve_handler in self.valve_handlers.values():
            if valve_handler.active or not only_active:
                tasks.append(valve_handler.finish())

        await asyncio.gather(*tasks)

    async def clear(self):
        """Cleanly finishes tasks and other clean-up tasks."""

        sshkeyboard.stop_listening()

        if self._progress_bar is not None:
            self._progress_bar.close()

        self._alerts_monitor_task = await cancel_task(self._alerts_monitor_task)

    @overload
    def fail(self, error: str, raise_error: bool = True) -> NoReturn: ...

    @overload
    def fail(self, error=None, raise_error: bool = True) -> None: ...

    def fail(
        self,
        error: str | None = None,
        raise_error: bool = True,
    ) -> NoReturn | None:
        """Sets the fail flag and event time."""

        self.failed = True
        self.event_times.fail_time = get_now()

        if error:
            self.error = error
            self.log.error(error)
            if raise_error:
                raise RuntimeError(error)

    @overload
    async def abort(
        self,
        error: str | None,
        close_valves: bool,
        raise_error: bool = True,
    ) -> NoReturn: ...

    @overload
    async def abort(
        self,
        error: str | None,
        close_valves: bool,
        raise_error: Literal[True] = True,
    ) -> NoReturn: ...

    @overload
    async def abort(
        self,
        error: str | None,
        close_valves: bool,
        raise_error: Literal[False] = False,
    ) -> None: ...

    async def abort(
        self,
        error: str | None = None,
        close_valves: bool = False,
        raise_error: bool = True,
    ) -> NoReturn | None:
        """Aborts the fill."""

        self.aborted = True
        self.event_times.abort_time = get_now()

        if close_valves:
            await self.close_valves(only_active=False)

        self.fail(error=error or "Aborted.", raise_error=raise_error)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-12-03
# @Filename: runner.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import itertools
import logging
import pathlib
import signal
import sys
from datetime import timedelta

from typing import TYPE_CHECKING

import httpx
import polars

from sdsstools.utils import run_in_executor

from lvmcryo.config import Actions
from lvmcryo.handlers import LN2Handler, close_all_valves
from lvmcryo.handlers.ln2 import get_now
from lvmcryo.notifier import Notifier


if TYPE_CHECKING:
    from lvmcryo.config import Config
    from lvmcryo.tools import DBHandler


__all__ = ["ln2_runner"]


async def signal_handler(handler: LN2Handler, log: logging.Logger):
    """Handles signals to close all valves and exit cleanly."""

    log.error("User aborted the process. Closing all valves before exiting.")
    await handler.close_valves(only_active=False)
    await handler.clear()

    handler.failed = True
    handler.aborted = True

    handler.event_times.fail_time = get_now()
    handler.event_times.abort_time = get_now()
    handler.event_times.end_time = get_now()

    log.error("Exiting now. Not data or notifications will be sent.")
    sys.exit(1)


async def ln2_runner(
    handler: LN2Handler,
    config: Config,
    notifier: Notifier | None = None,
    db_handler: DBHandler | None = None,
):
    """Runs the purge/fill process.

    This function is usually called by the CLI.

    """

    if notifier is None:
        # Create a notifier but immediately disable it.
        notifier = Notifier()
        notifier.disabled = True

    log = handler.log

    # Record options used. If in no_promp is False (the default), the full
    # configuration has already been printed for confirmation, so we only
    # save this with debug level.
    config_json = config.model_dump_json(indent=2)
    log.log(
        logging.INFO if config.no_prompt else logging.DEBUG,
        f"Running {config.action.value} with configuration:\n{config_json}",
    )

    if config.dry_run:
        log.warning("Running in dry-run mode. No valves will be operated.")

    # Inform of the start of the fill in Slack.
    now_str = get_now().strftime("%H:%M:%S")

    action = config.action.value

    if config.use_thermistors or config.purge_time is None or config.fill_time is None:
        lvmweb_url = config.internal_config["notifications.lvmweb_fill_url"]
        message = f"Starting LN₂ `{action}` at {now_str}."
        if db_handler and db_handler.pk is not None:
            fill_url = lvmweb_url.format(fill_id=db_handler.pk)
            message += f" Details can be followed from the <{fill_url}|LVM webapp>."

        await notifier.post_to_slack(message)

    else:
        await notifier.post_to_slack(
            f"Starting LN₂ `{action}` at {now_str} with "
            f"purge_time={config.purge_time} and "
            f"fill_time={config.fill_time}."
        )

    max_temperature = config.max_temperature if config.check_temperatures else None
    max_pressure = config.max_pressure if config.check_pressures else None
    await handler.check(
        max_pressure=max_pressure,
        max_temperature=max_temperature,
        check_thermistors=config.use_thermistors,
    )

    # Register signals that will trigger a valve shutdown and clean exit.
    for signame in ("SIGINT", "SIGTERM"):
        asyncio.get_running_loop().add_signal_handler(
            getattr(signal, signame),
            lambda: asyncio.create_task(signal_handler(handler, log)),
        )

    log.info(f"Closing all valves before {action}.")
    await close_all_valves(dry_run=config.dry_run)

    if handler.failed or handler.aborted:
        await handler.clear()
        raise RuntimeError("The LN2 handler was aborted before starting the fill.")

    if config.action == Actions.purge_fill or config.action == Actions.purge:
        await notifier.post_to_slack("Starting purge.")
        max_purge_time = config.purge_time or config.max_purge_time
        await handler.purge(
            use_thermistor=config.use_thermistors,
            min_purge_time=config.min_purge_time,
            max_purge_time=max_purge_time,
            prompt=not config.no_prompt,
            preopen_cb=db_handler.write if db_handler is not None else None,
        )

        if handler.failed or handler.aborted:
            await handler.clear()
            raise RuntimeError(handler.error or "Purge failed or was aborted.")

    if config.action == Actions.purge_fill or config.action == Actions.fill:
        await notifier.post_to_slack("Starting fill.")
        max_fill_time = config.fill_time or config.max_fill_time
        await handler.fill(
            use_thermistors=config.use_thermistors,
            require_all_thermistors=config.require_all_thermistors,
            min_fill_time=config.min_fill_time,
            max_fill_time=max_fill_time,
            prompt=not config.no_prompt,
            preopen_cb=db_handler.write if db_handler is not None else None,
        )

        if handler.failed or handler.aborted:
            await handler.clear()
            raise RuntimeError(handler.error or "Fill failed or was aborted.")

    await notifier.post_to_slack(f"LN₂ `{action}` completed successfully.")
    await handler.clear()

    handler.event_times.end_time = get_now()


async def post_fill_tasks(
    handler: LN2Handler,
    notifier: Notifier | None = None,
    write_data: bool = False,
    data_path: str | pathlib.Path | None = None,
    data_extra_time: float | None = None,
    api_data_route: str = "http://lvm-hub.lco.cl:8090/api/spectrographs/fills/measurements",
    generate_data_plots: bool = True,
) -> dict[str, pathlib.Path]:
    """Runs the post-fill tasks.

    Parameters
    ----------
    handler
        The `.LN2Handler` instance.
    write_data
        Whether to collect fill metrology data and write it to disk.
    data_path
        The path where to write the data. If `None` writes it to the current directory.
    data_extra_time
        Extra time to wait after the fill before collecting data.
    api_data_route
        The API route to retrive the fill data.
    generate_data_plots
        Whether to generate plots from the data.

    Returns
    -------
    plot_paths
        A mapping of plot type to path.

    """

    log = handler.log
    log.info("Running post-fill tasks.")

    event_times = handler.event_times
    plot_paths: dict[str, pathlib.Path] = {}

    if (
        not write_data
        or not event_times.start_time
        or not event_times.end_time
        or not api_data_route
    ):
        write_data = False
        log.debug("Skipping data collection.")

    if event_times.start_time and event_times.end_time:
        interval = (event_times.end_time - event_times.start_time).total_seconds()
        if interval < 120:
            log.warning("Fill duration was less than 2 minutes. Not collecting data.")
            write_data = False

    if write_data and event_times.start_time and event_times.end_time:
        if data_extra_time:
            if handler.aborted or handler.failed:
                log.warning(
                    "Fill was aborted or failed. Not waiting "
                    "additional time to collect data."
                )
            else:
                log.info(f"Waiting {data_extra_time} seconds before collecting data.")

                if notifier is not None:
                    await notifier.post_to_slack(
                        f"Fill notifications will be delayed {data_extra_time:.0f} "
                        "seconds while collecting post-fill data."
                    )

                await asyncio.sleep(data_extra_time)

        if data_path is None:
            data_path = pathlib.Path.cwd() / "fill_data.parquet"
        else:
            data_path = pathlib.Path(data_path)

        try:
            log.info("Retrieving and writing measurements.")

            end_time = event_times.end_time + timedelta(seconds=data_extra_time or 0.0)
            async with httpx.AsyncClient(follow_redirects=True) as client:
                response = await client.get(
                    api_data_route,
                    params={
                        "start_time": int(event_times.start_time.timestamp()),
                        "end_time": int(end_time.timestamp()),
                    },
                )
                response.raise_for_status()

                data = (
                    polars.DataFrame(response.json())
                    .with_columns(polars.col.time.cast(polars.Datetime("ms")))
                    .sort("time")
                    .drop_nulls()
                )

                data_path.parent.mkdir(parents=True, exist_ok=True)
                data.write_parquet(data_path)
                log.debug(f"Fill data written to {data_path!s}")

                if generate_data_plots:
                    log.debug("Generating plots.")
                    plot_path_root = str(data_path.with_suffix(""))
                    plot_paths = await run_in_executor(
                        generate_plots,
                        data,
                        plot_path_root,
                    )
                    plot_paths_transparent = await run_in_executor(
                        generate_plots,
                        data,
                        plot_path_root,
                        transparent=True,
                    )
                    plot_paths.update(plot_paths_transparent)
                    log.debug(f"Plots saved to {plot_path_root}*.")

        except Exception as ee:
            log.error(f"Failed to retrieve fill data from API: {ee!r}")

    return plot_paths


def generate_plots(
    data: polars.DataFrame,
    plot_path_root: str,
    transparent: bool = False,
    include_ccd_tempratures: bool = False,
):
    """Generates measurement plots.

    Parameters
    ----------
    data
        The data to plot.
    plot_path_root
        The root path where to save the plots. A number of plots will be generated
        using this root path.
    transparent
        Whether to save the plots with a transparent background. The colours will
        be adjusted to be visible on a dark background. The paths will be suffixed
        with ``_transparent``. No PDFs will be generated in this case.
    include_ccd_tempratures
        Whether to include CCD temperatures in the temperature plot.

    Returns
    -------
    paths
        A mapping of plot type to path.

    """

    import matplotlib.pyplot as plt

    paths: dict[str, pathlib.Path] = {}

    colours: dict[str, str] = {
        "r": "red",
        "b": "cyan" if transparent else "blue",
        "z": "magenta",
    }
    linestyles: dict[str, str] = {"1": "-", "2": "--", "3": "-."}

    transparent_suffix = "_transparent" if transparent else ""

    date = data[0, "time"].strftime("%Y-%m-%d")

    with plt.ioff():
        if transparent:
            plt.style.use("dark_background")
        else:
            plt.style.use("seaborn-v0_8-whitegrid")
            plt.rcParams["axes.facecolor"] = "white"
            plt.rcParams["figure.facecolor"] = "white"
            plt.rcParams["savefig.facecolor"] = "white"

        # Pressures
        fig, ax = plt.subplots(figsize=(12, 8))

        pressures = data.select(
            polars.col.time,
            polars.selectors.starts_with("pressure_"),
        )

        for spec, camera in itertools.product("123", "brz"):
            column = f"pressure_{camera}{spec}"
            if column not in pressures.columns:
                continue

            cam_pressure = pressures[column].to_numpy()

            colour = colours.get(camera, "w" if transparent else "k")
            linestyle = linestyles.get(spec, "-")

            label = f"{camera}{spec}"

            ax.plot(
                pressures["time"].to_numpy(),
                cam_pressure,
                label=label,
                color=colour,
                linestyle=linestyle,
            )

        ax.set_title(f"Pressure during fill — {date}")
        ax.set_xlabel("Time")
        ax.set_ylabel("Pressure [torr]")

        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), prop={"size": 9})

        plt.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))

        if not transparent:
            path = f"{plot_path_root}_pressure{transparent_suffix}.pdf"
            fig.savefig(path)
            paths[f"pressure{transparent_suffix}_pdf"] = pathlib.Path(path)

        path = f"{plot_path_root}_pressure{transparent_suffix}.png"
        fig.savefig(path, dpi=300, transparent=transparent)
        paths[f"pressure{transparent_suffix}_png"] = pathlib.Path(path)

        plt.close(fig)

        # Temperatures
        fig, ax = plt.subplots(figsize=(12, 8))

        temps = data.select(
            polars.col.time,
            polars.selectors.starts_with("temp_"),
        )

        for spec, camera, sensor in itertools.product("123", "brz", ["ln2", "ccd"]):
            if sensor == "ccd" and not include_ccd_tempratures:
                continue

            column = f"temp_{camera}{spec}_{sensor}"
            if column not in temps.columns:
                continue

            cam_temp = temps[column].to_numpy()

            colour = colours.get(camera, "w" if transparent else "k")
            linestyle = linestyles.get(spec, "-")
            linewidth = 1.5 if sensor == "ln2" else 1

            label = f"{camera}{spec} ({sensor.upper()})"

            ax.plot(
                temps["time"].to_numpy(),
                cam_temp,
                label=label,
                color=colour,
                linestyle=linestyle,
                linewidth=linewidth,
            )

        ax.set_title(f"Temperature during fill — {date}")
        ax.set_xlabel("Time")
        ax.set_ylabel("Temperature [C]")

        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), prop={"size": 9})

        if not transparent:
            path = f"{plot_path_root}_temps{transparent_suffix}.pdf"
            fig.savefig(path)
            paths[f"temps{transparent_suffix}_pdf"] = pathlib.Path(path)

        path = f"{plot_path_root}_temps{transparent_suffix}.png"
        fig.savefig(path, dpi=300, transparent=transparent)
        paths[f"temps{transparent_suffix}_png"] = pathlib.Path(path)

        plt.close(fig)

        # Thermistors
        fig, ax = plt.subplots(figsize=(12, 8))

        therms = data.select(
            polars.col.time,
            polars.selectors.starts_with("thermistor_"),
        )

        cameras = ["".join(item)[::-1] for item in itertools.product("123", "brz")]
        for channel in ["supply"] + cameras:
            column = f"thermistor_{channel}"
            if column not in therms.columns:
                continue

            cam_therm = therms[column].to_numpy()

            if len(channel) == 2:
                camera, spec = channel
                colour = colours.get(camera, "w" if transparent else "k")
                linestyle = linestyles.get(spec, "-")
            else:
                colour = "g" if transparent else "k"
                linestyle = "-"

            ax.plot(
                therms["time"].to_numpy(),
                cam_therm,
                label=channel,
                color=colour,
                linestyle=linestyle,
            )

        ax.set_title(f"Thermistors during fill — {date}")
        ax.set_xlabel("Time")
        ax.set_ylabel("State")

        ax.legend(loc="center left", bbox_to_anchor=(1, 0.5), prop={"size": 9})

        if not transparent:
            path = f"{plot_path_root}_thermistors{transparent_suffix}.pdf"
            fig.savefig(path)
            paths[f"thermistors{transparent_suffix}_pdf"] = pathlib.Path(path)

        path = f"{plot_path_root}_thermistors{transparent_suffix}.png"
        fig.savefig(path, dpi=300, transparent=transparent)
        paths[f"thermistors{transparent_suffix}_png"] = pathlib.Path(path)

        plt.close(fig)

    return paths



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-09-14
# @Filename: config.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import datetime
import pathlib
import warnings
from enum import Enum
from functools import lru_cache

from typing import Annotated, Any, Self

from click.core import ParameterSource
from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    field_serializer,
    field_validator,
    model_validator,
)

from sdsstools.configuration import Configuration, read_yaml_file


class Actions(str, Enum):
    purge_fill = "purge-and-fill"
    purge = "purge"
    fill = "fill"
    abort = "abort"
    clear_lock = "clear-lock"


class InteractiveMode(str, Enum):
    auto = "auto"
    yes = "yes"
    no = "no"


class NotificationLevel(str, Enum):
    info = "info"
    error = "error"


class ThermistorConfig(BaseModel):
    """Thermistor configuration model."""

    channel: str | None = None
    monitoring_interval: float = 1.0
    close_valve: bool = True
    required_active_time: float = 10.0
    disabled: bool = False


class ValveConfig(BaseModel):
    """Valve configuration model."""

    actor: str
    outlet: str
    thermistor: ThermistorConfig | None = Field(default_factory=ThermistorConfig)

    @model_validator(mode="after")
    def validate_after(self) -> Self:
        """Validates the valve configuration after the fields have been set."""

        if self.thermistor and self.thermistor.channel is None:
            self.thermistor.channel = self.outlet

        return self


ExcludedField = Field(repr=False, exclude=True)


@lru_cache()
def get_internal_config(path: pathlib.Path | str | None = None) -> Configuration:
    """Returns the internal configuration."""

    default_path = pathlib.Path(__file__).parent / "config.yaml"

    if path is None:
        return read_yaml_file(default_path)

    return Configuration(path, base_config=default_path)


class Config(BaseModel):
    """Configuration model.

    This model defines the options with which a purge/fill process will run.
    It is usually generated from the options passed to a CLI command after
    they are parsed by `.process_cli_options`.

    """

    action: Actions = Actions.purge_fill
    cameras: list[str] = []  # Empty list means all cameras.

    interactive: InteractiveMode = InteractiveMode.auto
    no_prompt: bool = False
    dry_run: bool = False
    clear_lock: bool = False
    with_traceback: bool = False

    use_thermistors: bool = True
    require_all_thermistors: bool = False
    check_pressures: bool = True
    check_temperatures: bool = True
    max_pressure: float | None = None
    max_temperature: float | None = None
    purge_time: float | None = None
    min_purge_time: float | None = None
    max_purge_time: float | None = None
    fill_time: float | None = None
    min_fill_time: float | None = None
    max_fill_time: float | None = None

    verbose: bool = False
    quiet: bool = False
    notify: bool = True
    slack: bool = True
    email: bool = True
    email_level: NotificationLevel = NotificationLevel.error

    write_log: bool = False
    log_path: pathlib.Path | None = None
    write_json: bool = False
    write_data: bool = False
    data_path: pathlib.Path | None = None
    data_extra_time: float = 0.0

    version: str | None = None

    valve_info: Annotated[dict[str, ValveConfig], ExcludedField] = {}

    config_file: Annotated[pathlib.Path | None, ExcludedField] = None
    _internal_config: Annotated[Configuration, PrivateAttr] = Configuration({})

    error: Annotated[bool, ExcludedField] = False

    profile: str | None = None
    param_source: Annotated[dict[str, ParameterSource | None], ExcludedField] = {}

    def model_post_init(self, __context: Any) -> None:
        self._internal_config = get_internal_config(self.config_file)
        return super().model_post_init(__context)

    @property
    def internal_config(self):
        return self._internal_config

    @field_validator("interactive", mode="after")
    @classmethod
    def validate_interactive(cls, value: InteractiveMode) -> InteractiveMode:
        """Returns the interactive mode."""

        from lvmcryo.tools import is_container

        is_container = is_container()

        if value in ["yes", True, 1]:
            if is_container:
                warnings.warn("Interactive mode may not work in containers.")
            return InteractiveMode.yes

        if value in ["no", False, 0]:
            return InteractiveMode.no

        if value in ["auto", None]:
            if is_container:
                return InteractiveMode.no
            else:
                return InteractiveMode.yes

        raise ValueError(f"Invalid value for interactive mode: {value!r}")

    @model_validator(mode="before")
    @classmethod
    def before_validator(cls, data: Any) -> Any:
        if not isinstance(data, dict):
            # Not sure if this is likely to happen.
            return data

        # Ensure we have an internal configuration dictionary.
        config = get_internal_config(data.get("config_file", None))
        defaults = config.get("defaults", {})

        # If a profile has been passed, we update the input parameters with those
        # in the profile. But we want to do that only for cases in which the user
        # has not explicitly passed the parameter as a flag.
        if (profile := data["profile"]) is not None:
            profile_data = config.get("profiles", {}).get(profile, {})
            param_source = data.get("param_source", {})
            for key in profile_data:
                if key not in data:
                    warnings.warn(f"Unknwon parameter in profile: {key}")
                    continue
                psource = param_source.get(key, None)
                if psource != ParameterSource.COMMANDLINE:
                    data[key] = profile_data[key]

        # Use internal configuration files to fill in missing fields.
        for field in [
            "min_purge_time",
            "max_purge_time",
            "min_fill_time",
            "max_fill_time",
            "max_pressure",
            "max_temperature",
        ]:
            default_value = defaults.get(field, None)
            if field not in data or data[field] is None:
                data[field] = default_value

        if "cameras" not in data or data["cameras"] == []:
            if "cameras" not in defaults:
                raise ValueError("No cameras defined in the configuration file.")
            data["cameras"] = defaults["cameras"]

        if "valve_info" not in data or data["valve_info"] is None:
            if "valve_info" not in config:
                raise ValueError("No valve_info defined in the configuration file.")
            data["valve_info"] = config["valve_info"]

        # Fill out the interactive value for now with None. Will be validated later.
        if "interactive" not in data:
            data["interactive"] = None

        return data

    @field_serializer("action", "interactive", "email_level")
    def serialize_enums(self, enum) -> str:
        """Serialises the enumerations to string."""

        return enum.value

    @field_serializer("log_path", "data_path")
    def serialize_path(self, path: pathlib.Path | None) -> str | None:
        """Serialises the path."""

        return str(path) if path is not None else None

    @model_validator(mode="after")
    def validate_after(self) -> Self:
        """Performs validations after the fields have been set."""

        # At this point we should have narrowed down the interactive mode.
        if self.interactive not in [InteractiveMode.yes, InteractiveMode.no]:
            raise ValueError("Invalid interactive mode after validation.")

        # If we are not in interactive mode, we set no_prompt to True.
        if self.interactive == InteractiveMode.no:
            self.no_prompt = True

        defaults = self._internal_config.get("defaults", {})

        # Ensure the log path is set correctly.
        if self.log_path is not None:
            self.write_log = True
        elif self.write_log is False:
            self.log_path = None
        else:
            if self.log_path is not None:
                self._check_log_path(self.log_path)
            else:
                default_path = defaults.get("log_path", "./{timestamp}.log")
                self.log_path = pathlib.Path(
                    default_path.format(timestamp=self._log_basename())
                )

        # Ensure the data path is set correctly.
        if self.data_path is not None:
            self.write_data = True
        elif self.write_data is False:
            self.data_path = None
        else:
            if self.data_path is not None:
                self._check_log_path(self.data_path)
            else:
                if self.log_path is not None:
                    self.data_path = self.log_path.with_suffix(".parquet")
                else:
                    default_path = defaults.get("data_path", "./{timestamp}.parquet")
                    self.data_path = pathlib.Path(
                        default_path.format(timestamp=self._log_basename("parquet"))
                    )

        # We won't write a JSON file if we are not writing a normal log.
        if not self.write_log:
            self.write_json = False

        # We cannot purge/fill without thermistors and without fill and purge times
        # unless we are in interactive mode.
        if not self.use_thermistors and not self.interactive:
            no_purge_time = self.purge_time is None
            no_fill_time = self.fill_time is None

            if self.action == Actions.purge_fill and (no_fill_time or no_purge_time):
                raise ValueError(
                    "use_thermistors=False requires interactive mode "
                    "or defining purge and fill times."
                )

            if self.action == Actions.purge and no_purge_time:
                raise ValueError(
                    "use_thermistors=False requires interactive mode "
                    "or defining a purge time."
                )

            if self.action == Actions.fill and no_fill_time:
                raise ValueError(
                    "use_thermistors=False requires interactive mode "
                    "or defining a fill time."
                )

        # use_thermistors cannot be used with fill/purge times.
        if self.use_thermistors and (self.purge_time or self.fill_time):
            raise ValueError("use_thermistors cannot be used with purge/fill times.")

        # If we are using thermistors, require max_fill_time and max_purge_time.
        if self.use_thermistors:
            if self.max_fill_time is None or self.max_purge_time is None:
                raise ValueError(
                    "use_thermistors=True defining max_fill_time and max_purge_time. "
                    "This probably indicates a problem with your configuration file."
                )

        return self

    def _check_log_path(self, path: pathlib.Path, extension="log") -> pathlib.Path:
        """Checks that the log path is valid."""

        if path.is_dir():
            return path / self._log_basename(extension)
        else:
            if path.exists():
                raise ValueError(f"Path {path} already exists.")

        return path

    def _log_basename(self, extension: str | None = None) -> str:
        """Generates a log basename with a timestamp."""

        extension = extension or ""

        now = datetime.datetime.now(datetime.UTC)
        iso = now.isoformat(timespec="seconds").split("+")[0]

        if extension and not extension.startswith("."):
            extension = f".{extension}"

        return f"lvmcryo_{iso}{extension}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-10-08
# @Filename: validate.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import logging
import pathlib
from functools import partial

from typing import NoReturn

import polars

from sdsstools.logger import SDSSLogger

from lvmcryo.handlers.ln2 import LN2Handler
from lvmcryo.tools import get_fake_logger


__all__ = ["validate_fill"]


def log_or_raise(
    log: logging.Logger | SDSSLogger | None,
    raise_on_error: bool,
    message: str,
    level: int = logging.INFO,
):
    """Logs a message or raises an exception.

    Parameters
    ----------
    log
        A logger to output messages to the user.
    message
        The message to log or raise.
    level
        The logging level to use if ``log`` is provided.
    raise_on_error
        If `True`, raises an exception if the message level is ``ERROR`` or higher.

    """

    log = log or get_fake_logger()

    if level >= logging.ERROR and raise_on_error:
        raise RuntimeError(message)

    log.log(level, message)


def validate_fill(
    ln2_handler: LN2Handler,
    data: polars.DataFrame | pathlib.Path | str,
    log: logging.Logger | SDSSLogger | None = None,
    raise_on_error: bool = False,
) -> tuple[bool, str | None] | NoReturn:
    """Validates post-fill data.

    Parameters
    ----------
    ln2_handler
        The LN2 handler.
    data
        A dataframe with the data to validate. Usually the output of
        `.runner.post_fill`.
    log
        A logger to output messages to the user.
    raise_on_error
        If `True`, raises an exception if the validation fails. Otherwise the
        validation errors are logged and the result is returned as a boolean.

    Returns
    -------
    result
        A tuple in which the first element is a boolean indicating whether the
        validation failed, and the second element is a message with the validation
        result.

    """

    log_p = partial(log_or_raise, log, raise_on_error)
    log_p("Validating post-fill data.")

    if isinstance(data, (pathlib.Path, str)):
        file_ = pathlib.Path(data)
        if not file_.exists():
            log_p(
                f"File {file_} does not exist. Cannot validate fill.",
                level=logging.WARNING,
            )
            return (False, None)

        data = polars.read_parquet(data)

    event_times = ln2_handler.event_times

    failed: bool = False
    error: str | None = None

    if data.height == 0:
        log_p("No post-fill data was collected.", level=logging.ERROR)
        return (False, "No post-fill data was collected.")

    # Make sure the time is UTC.
    data = data.with_columns(polars.col.time.dt.replace_time_zone("UTC"))

    # 1. Check that the LN2 temperature after the fill is lower than before.
    # We only do this if we have collected enough data after the fill completes.

    if event_times.fill_start is None or event_times.fill_complete is None:
        # There was no LN2 fill.
        pass
    else:
        ln2_temp = data.select(polars.col.time, polars.col("^temp_[rbz][1-3]_ln2$"))
        ln2_temp = ln2_temp.sort(polars.col.time)

        # Check that the last point was taken at least 3 minutes after the fill.
        max_time = ln2_temp["time"].max()
        extra_time = (max_time - event_times.end_time).total_seconds()  # type: ignore
        if extra_time < 3 * 60:
            log_p(
                "Not enough data collected after the fill "
                "to check the LN2 temperature difference.",
                level=logging.WARNING,
            )
        else:
            for column in ln2_temp.columns:
                if column == "time":
                    continue

                camera = column.split("_")[1]

                temp0 = ln2_temp[0, column]
                temp1 = ln2_temp[-1, column]

                if temp1 > temp0:
                    failed = True
                    error = (
                        f"LN2 temperature for camera {camera} increased "
                        f"from {temp0:.2f} to {temp1:.2f} degC after the fill."
                    )
                    log_p(error, level=logging.ERROR)
                    break

    return (failed, error)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-12-03
# @Filename: notifier.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib
import re
import smtplib
import traceback
import uuid
import warnings
from datetime import datetime
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from typing import TYPE_CHECKING

import httpx
import pygments
from pydantic import BaseModel, ValidationError
from pygments.formatters import HtmlFormatter
from pygments.lexers.python import PythonTracebackLexer

from lvmopstools.devices.specs import spectrograph_pressures, spectrograph_temperatures

from lvmcryo.config import NotificationLevel, get_internal_config
from lvmcryo.tools import get_fake_logger, render_template


if TYPE_CHECKING:
    from lvmcryo.handlers import LN2Handler


GRAFANA_URL = "https://lvm-grafana.lco.cl/d/ec97aef8-071f-4f9f-9cc9-2f3f206d8308/cryostat-temperatures-and-pressures?orgId=1&refresh=10s"


class NotifierConfig(BaseModel):
    """Configuration for the Notifier class."""

    api_route: str
    slack_channel: str
    slack_from: str = "LN₂ Helper"

    email_recipients: list[str]
    email_server: str
    email_from: str
    email_reply_to: str | None = None

    lvmweb_fill_url: str


class Notifier:
    """Sends notifications over Slack or email."""

    def __init__(self, config_data: dict | str | pathlib.Path | None = None):
        if config_data is None or isinstance(config_data, (str, pathlib.Path)):
            config_data = get_internal_config(config_data)

        if not config_data["notifications"]:
            raise ValidationError("Configuration does not have notifications section.")

        self.config = NotifierConfig(
            api_route=config_data["api_routes"]["create_notification"],
            **config_data["notifications"],
        )

        self.disabled: bool = False
        self.slack_disabled: bool = False
        self.email_disabled: bool = False

    def __repr__(self):
        return f"<Notifier (disabled={str(self.disabled).lower()})>"

    async def post_to_slack(
        self,
        text: str | None = None,
        level: NotificationLevel = NotificationLevel.info,
        channel: str | None = None,
    ):
        """Posts a message to Slack.

        Parameters
        ----------
        text
            Plain text to send to the Slack channel.
        level
            The level of the message. Determines the channel where the message
            is sent.
        channel
            The channel in the SSDS-V workspace where to send the message. Defaults
            to the configuration value.

        """

        if self.disabled or self.slack_disabled:
            return

        route = self.config.api_route

        channel = channel or self.config.slack_channel
        notification_level = "INFO" if level == NotificationLevel.info else "CRITICAL"

        try:
            async with httpx.AsyncClient(follow_redirects=True) as client:
                response = await client.post(
                    route,
                    json={
                        "message": text,
                        "level": notification_level,
                        "slack_channel": channel,
                        "email_on_critical": False,  # We handle our own mailing
                        "slack_extra_params": {"username": self.config.slack_from},
                    },
                )

            if response.status_code != 200:
                raise RuntimeError(response.text)

        except Exception as ee:
            warnings.warn(f"Failed sending message to Slack: {ee}")
            return False

        return True

    def send_email(
        self,
        html_message: str | None = None,
        plaintext_message: str | None = None,
        subject: str = "A message from lvmcryo",
        recipients: list[str] | None = None,
        from_address: str | None = None,
        images: dict[str, pathlib.Path | None] = {},
        email_server: str | None = None,
    ):
        """Sends an email to a list of recipients.

        Parameters
        ----------
        html_message
            The body of the email as an HTML string.
        plaintext_message
            An alternative plaintext version of the email body.
        subject
            The subject of the email.
        recipients
            A list of email addresses to send the email to. If ``None``, uses the
            default recipients in the configuration.
        from_address
            The email address to send the email from. If ``None``, uses the default
            address in the configuration.
        images
            A mapping of images to embed. This requires an ``html_message``. The
            mapping must be in the form ``{xyx: path}``, where ``xyz`` is the
            content ID key associated to the image. In the HTML text there must
            be a corresponding ``<img src="cid:xyz">`` tag.
        email_server
            The SMTP server to use. If ``None``, uses the default server in the
            configuration.

        """

        if self.disabled or self.email_disabled:
            return

        recipients = recipients or self.config.email_recipients
        from_address = from_address or self.config.email_from

        email_server = email_server or self.config.email_server
        email_host, *email_rest = email_server.split(":")
        email_port: int = 0
        if len(email_rest) == 1:
            email_port = int(email_rest[0])

        email_reply_to = self.config.email_reply_to or from_address

        msg = MIMEMultipart("alternative" if html_message else "mixed")
        msg["Subject"] = subject
        msg["From"] = from_address
        msg["To"] = ", ".join(recipients)
        msg["Reply-To"] = email_reply_to

        plain = MIMEText(
            plaintext_message
            or "A message was sent from lvmcryo but your email "
            "client cannot process HTML.",
            "plain",
        )
        msg.attach(plain)

        if html_message:
            for cid, image in images.items():
                # Need to replace the cid with a globally unique one.
                cid_unique = f"{cid}-{uuid.uuid4()!s}"
                html_message = html_message.replace(f"cid:{cid}", f"cid:{cid_unique}")

                if image is not None and pathlib.Path(image).exists():
                    with open(image, "rb") as fp:
                        image = MIMEImage(fp.read())

                    # Specify the  ID according to the img src in the HTML part
                    image.add_header("Content-ID", f"<{cid_unique}>")
                    msg.attach(image)

            html = MIMEText(html_message, "html")
            msg.attach(html)

        try:
            with smtplib.SMTP(host=email_host, port=email_port) as smtp:
                smtp.sendmail(from_address, ", ".join(recipients), msg.as_string())
        except Exception as ee:
            warnings.warn(f"Failed sending email: {ee}")
            return False

        return True

    async def notify_after_fill(
        self,
        success: bool,
        error_message: str | Exception | None = None,
        handler: LN2Handler | None = None,
        post_to_slack: bool = True,
        send_email: bool = True,
        include_status: bool = True,
        include_log: bool = True,
        images: dict[str, pathlib.Path | None] = {},
        record_pk: int | None = None,
    ):
        """Notifies a fill success or failure to the users."""

        log = handler.log if handler else get_fake_logger()

        spec_data: dict[str, dict] | None = None
        if include_status:
            try:
                temp_data = await spectrograph_temperatures()
                pressure_data = await spectrograph_pressures()

                # Massage the data into the format that the template expects.
                cryostats = list(pressure_data)
                spec_data = {}
                for cryostat in cryostats:
                    spec_data[cryostat] = {
                        "ccd": temp_data.get(f"{cryostat}_ccd", -999),
                        "ln2": temp_data.get(f"{cryostat}_ln2", -999),
                        "pressure": pressure_data.get(cryostat, -999),
                    }

            except Exception as err:
                log.error(f"Failed retrieving spectrograph data: {err!r}")

        formatter = HtmlFormatter(style="default")

        log_lines = []
        if include_log:
            fh = getattr(log, "fh", None)
            if fh:
                fh.flush()
            log_filename = getattr(log, "log_filename", None)
            if log_filename:
                log_lines_raw = open(log_filename, "r").readlines()

                # Remove the milliseconds from the log lines.
                pattern = re.compile(r"^((?:\d+-?)+ (?:\d{2}:?)+),\d{3}( - )")
                for line in log_lines_raw:
                    new_line = pattern.sub(r"\1\2", line)
                    log_lines.append(new_line)

        if post_to_slack:
            try:
                if success:
                    slack_message = "LN₂ fill completed successfully."
                else:
                    slack_message = (
                        "Something went wrong with the LN₂ fill. "
                        "Please check the status of the spectrographs. "
                        f"Grafana plots are available <{GRAFANA_URL}|here>."
                    )
                    if error_message:
                        if isinstance(error_message, Exception):
                            trace = "".join(traceback.format_exception(error_message))
                            slack_message += f"\n```{trace}```"
                        else:
                            slack_message += f"\nThe error was: {error_message}"

                await self.post_to_slack(
                    text=slack_message,
                    level=NotificationLevel.error,
                )

            except Exception as err:
                log.error(f"Failed posting to Slack: {err!r}")

        lvmweb_url: str | None = None
        if record_pk is not None:
            lvmweb_url = self.config.lvmweb_fill_url.format(fill_id=record_pk)
            await self.post_to_slack(
                "Information about the fill can be found at "
                f"<{lvmweb_url}|this link>."
            )

        if send_email:
            subject: str = (
                "SUCCESS: LVM LN2 fill completed"
                if success
                else "ERROR: LVM LN2 fill failed"
            )
            plain_message: str = (
                "LN2 fill completed successfully."
                if success
                else "LN2 fill failed. Please check the cryostats."
            )
            html_message: str | None = None
            email_error: str | None = None
            template = "success_message.html" if success else "alert_message.html"
            valve_data = {}

            try:
                try:
                    if not success:
                        if isinstance(error_message, Exception):
                            email_error = pygments.highlight(
                                "".join(traceback.format_exception(error_message)),
                                PythonTracebackLexer(),
                                formatter,
                            )
                        else:
                            email_error = error_message

                        plain_message = (
                            "LN2 fill failed. Please check the cryostats. "
                            "You are not receiving a full message because your email "
                            "client does not support HTML."
                        )

                    if handler:
                        valve_times = handler.get_valve_times()
                        for valve in valve_times:
                            valve_data[valve] = {
                                "open_time": None,
                                "close_time": None,
                                "elapsed": None,
                                "thermistor_after": None,
                                "timed_out": False,
                            }

                            open_time = valve_times[valve]["open_time"]
                            close_time = valve_times[valve]["close_time"]
                            th_first = valve_times[valve]["thermistor_first_active"]

                            if open_time is None or close_time is None:
                                continue

                            assert isinstance(open_time, datetime)
                            assert isinstance(close_time, datetime)

                            delta = close_time - open_time

                            th_after: float | None = None
                            if th_first is not None:
                                assert isinstance(th_first, datetime)
                                delta_first = th_first - open_time
                                th_after = delta_first.total_seconds()

                            valve_data[valve] = {
                                "open_time": open_time,
                                "close_time": close_time,
                                "elapsed": delta.total_seconds(),
                                "thermistor_after": th_after,
                                "timed_out": valve_times[valve]["timed_out"],
                            }

                    has_images = any([image is not None for image in images.values()])
                    html_message = render_template(
                        template,
                        render_data=dict(
                            event_times=handler.event_times if handler else {},
                            spec_data=spec_data,
                            log_lines=log_lines if len(log_lines) > 0 else None,
                            log_css=formatter.get_style_defs(),
                            error=email_error,
                            has_images=has_images,
                            valve_data=valve_data,
                            grafana_url=GRAFANA_URL,
                            lvmweb_url=lvmweb_url,
                        ),
                    )

                except Exception as err:
                    log.error(f"Failed rendering template: {err!r}")
                    log.warning("Sending a plain text message.")
                    html_message = None

                self.send_email(
                    subject=subject,
                    html_message=html_message,
                    plaintext_message=plain_message,
                    images=images,
                )

            except Exception as err:
                log.error(f"Failed sending email: {err!r}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-11-10
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from sdsstools import get_package_version


# pip package name
NAME = "lvmcryo"


# package name should be pip package name
__version__ = get_package_version(path=__file__, package_name=NAME)
